==> herrie-0.1.tar.gz: herrie-0.1/src/gui/gui_playq.c <==
/*
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: gui_playq.c 477 2006-08-01 06:51:34Z ed $
 */
/**
 * @file gui/gui_playq.c
 */

#include "config.h"

#include <assert.h>
#include <curses.h>

#include <gui/gui.h>
#include <playq/playq.h>

static WINDOW *win_playq;

static struct playq_entry *pe_top, *pe_selected;
static int idx_top, idx_selected;

void
gui_playq_init(void)
{
	win_playq = newwin(gui_size_playq_height, 0, 1, 0);
	wbkgdset(win_playq, COLOR_PAIR(GUI_COLOR_BLOCK));

	pe_top = pe_selected = NULL;
	idx_top = idx_selected = -1;
	gui_playq_refresh();
}

void
gui_playq_destroy(void)
{
	delwin(win_playq);
}

/**
 * @brief Scroll the viewport up if the cursor is too far up.
 */
static void
gui_playq_cursor_adjust_top(void)
{
	if (idx_top > idx_selected) {
		pe_top = pe_selected;
		idx_top = idx_selected;
	}
}

/**
 * @brief Scroll the viewport down if the cursor is too far down.
 */
static void
gui_playq_cursor_adjust_bottom(void)
{
	int i, idx;
	struct playq_entry *pe;

	pe = pe_selected;
	idx = idx_selected;

	/* Don't deal with empty selections */
	if (pe == NULL)
		return;
	
	for (i = 1; i < gui_size_playq_height; i++) {
		pe = playq_entry_prev(pe);
		idx--;

		/* The top of the list is in screen */
		if ((pe == NULL) || (pe == pe_top))
			return;
	}

	/* We've gone too far */
	pe_top = pe;
	idx_top = idx;
}

/**
 * @brief Calculate what width the column with the index numbers should
 *        have, based on the amount of songs in the playlist.
 */
static int
gui_playq_idxcol_width(void)
{
	int idx, len;

	/*
	 * Calculate the number of digits needed to print the number of
	 * entries
	 */
	for (len = 1, idx = playq_len; idx >= 10; len++, idx /= 10);

	return (len);
}

/**
 * @brief Fill playlist window with contents by acquiring the GUI lock
 *        and assuming that the playlist is already locked.
 */
static void
gui_playq_redraw(void)
{
	int i, idx, idxw, idxmaxw;
	char num[16];
	struct playq_entry *pe;

	GUI_LOCK;
	werase(win_playq);

	pe = pe_top;
	idx = idx_top;
	idxmaxw = gui_playq_idxcol_width();
	for (i = 0; i < gui_size_playq_height; i++) {
		if (pe == NULL) {
			/* We've reached the bottom of the list */
			mvwaddch(win_playq, i, 0, ' ');
			wbkgdset(win_playq, COLOR_PAIR(GUI_COLOR_BLOCK));
			wclrtobot(win_playq);
			break;
		}

		if (pe == pe_selected) {
			if (gui_input_curfocus == GUI_FOCUS_PLAYQ)
				wbkgdset(win_playq,
				    COLOR_PAIR(GUI_COLOR_SELECT));
			else
				wbkgdset(win_playq,
				    COLOR_PAIR(GUI_COLOR_DESELECT));

			assert(idx == idx_selected);
		}

		/* Small whitespace on the left, or > when black & white */
		if ((pe == pe_selected) && !has_colors()) {
			if (gui_input_curfocus == GUI_FOCUS_PLAYQ)
				mvwaddch(win_playq, i, 0, '>');
			else
				mvwaddch(win_playq, i, 0, '-');
		} else {
			mvwaddch(win_playq, i, 0, ' ');
		}
		wclrtoeol(win_playq);

		idxw = snprintf(num, sizeof num, "%d.", idx + 1);
		mvwaddstr(win_playq, i, 2 + idxmaxw - idxw, num);
		mvwaddstr(win_playq, i, idxmaxw + 3, pe->name);

		wbkgdset(win_playq, COLOR_PAIR(GUI_COLOR_BLOCK));

		pe = playq_entry_next(pe);
		idx++;
	}

	wrefresh(win_playq);

	GUI_UNLOCK;
}

#ifdef HAVE_WRESIZE
void
gui_playq_resize(void)
{
	wresize(win_playq, gui_size_playq_height, COLS);
	clearok(win_playq, TRUE);
	
	PLAYQ_LOCK;
	gui_playq_cursor_adjust_bottom();
	gui_playq_redraw();
	PLAYQ_UNLOCK;
}
#endif /* HAVE_WRESIZE */

void
gui_playq_refresh(void)
{
	PLAYQ_LOCK;
	gui_playq_redraw();
	PLAYQ_UNLOCK;
}

/**
 * @brief Adjust a reference with index to an entry in the playlist,
 *        right before another entry in the playlist gets removed.
 */
static void
gui_playq_adjust_pre_removal(struct playq_entry **pe,
    int *curidx, int newidx, int follow)
{
	struct playq_entry *pe_new;
	assert((*pe != NULL) && (*curidx >= 0));

	if (*curidx > newidx) {
		/* Item got removed above us */
		pe_new = playq_entry_next(*pe);
		if (follow || pe_new == NULL) {
			/* Follow our selection */
			(*curidx)--;
		} else {
			/* Keep the current offset */
			*pe = pe_new;
		}
	} else if (*curidx == newidx) {
		/* Current item gets removed */
		pe_new = playq_entry_next(*pe);
		if (pe_new != NULL) {
			/* Stay at the same position */
			*pe = pe_new;
		} else {
			/* Cannot go down - go up */
			*pe = playq_entry_prev(*pe);
			(*curidx)--;
		}
	}

	assert((*pe == NULL) == (*curidx < 0));
}

void
gui_playq_notify_pre_removal(int index)
{
	gui_playq_adjust_pre_removal(&pe_top, &idx_top, index, 0);
	gui_playq_adjust_pre_removal(&pe_selected, &idx_selected, index, 1);
	gui_playq_cursor_adjust_top();
}

/**
 * @brief Adjust a reference with index to an entry in the playlist,
 *        right after a new entry in the playlist gets inserted.
 */
static void
gui_playq_adjust_post_insertion(struct playq_entry **pe,
    int *curidx, int newidx, int follow)
{
	assert((*pe == NULL) == (*curidx < 0));

	if (*curidx < 0) {
		/* List was empty - we take the first slot */
		*pe = playq_entry_first();
		*curidx = 0;
	} else if (*curidx >= newidx) {
		/* It happened above us */
		if (follow) {
			/* Follow our selection */
			(*curidx)++;
		} else {
			/* Keep the current offset */
			*pe = playq_entry_prev(*pe);
		}
	}

	assert((*pe != NULL) && (*curidx >= 0));
}

void
gui_playq_notify_post_insertion(int index)
{
	gui_playq_adjust_post_insertion(&pe_top, &idx_top, index, 0);
	gui_playq_adjust_post_insertion(&pe_selected, &idx_selected, index, 1);
	gui_playq_cursor_adjust_bottom();
}

void
gui_playq_notify_post_randomization(void)
{
	int idx;

	/*
	 * After a randomization the indices do not match the
	 * corresponding objects. We could trace the objects and update
	 * the indices, but we might as well better keep the current
	 * indices and update the objects. This is more convenient for
	 * the user.
	 */
	if (idx_top >= 0) {
		/* Top position */
		for (idx = 0, pe_top = playq_entry_first();
		    idx < idx_top;
		    idx++, pe_top = playq_entry_next(pe_top));
		/* Selection - just continue previous traversal */
		for (pe_selected = pe_top;
		    idx < idx_selected;
		    idx++, pe_selected = playq_entry_next(pe_selected));
	}
}

void
gui_playq_notify_done(void)
{
	gui_playq_redraw();
}

void
gui_playq_cursor_up(void)
{
	struct playq_entry *pe;

	PLAYQ_LOCK;
	if (pe_selected != NULL) {
		pe = playq_entry_prev(pe_selected);
		if (pe != NULL) {
			/* XXX */
			if (pe_top == pe_selected) {
				pe_top = pe;
				idx_top--;
			}
			
			pe_selected = pe;
			idx_selected--;

			gui_playq_redraw();
		}
	}
	PLAYQ_UNLOCK;
}

void
gui_playq_cursor_down(void)
{
	struct playq_entry *pe;

	PLAYQ_LOCK;
	if (pe_selected != NULL) {
		pe = playq_entry_next(pe_selected);
		if (pe != NULL) {
			pe_selected = pe;
			idx_selected++;

			/*
			 * Make sure the cursoer doesn't leave the
			 * bottom of the screen
			 */
			gui_playq_cursor_adjust_bottom();
			gui_playq_redraw();
		}
	}
	PLAYQ_UNLOCK;
}

void
gui_playq_cursor_pageup(void)
{
	int i;

	PLAYQ_LOCK;
	if (idx_top <= 0)
		goto done;

	/* Cursor = bottom of the screen */
	pe_selected = playq_entry_next(pe_top);
	idx_selected = idx_top + 1;

	/* Adjust the viewport */
	for (i = 2; i < gui_size_playq_height && pe_top != NULL; i++) {
		pe_top = playq_entry_prev(pe_top);
		idx_top--;
	}
	if (pe_top == NULL) {
		pe_top = playq_entry_first();
		idx_top = 0;
	}
	
	/* Original screen may have had one item */
	if (pe_selected == NULL) {
		pe_selected = pe_top;
		idx_selected = idx_top;
	}
	
	gui_playq_redraw();
done:
	PLAYQ_UNLOCK;
}

void
gui_playq_cursor_pagedown(void)
{
	int i;

	PLAYQ_LOCK;
	/* A la vi: page down minus two */
	for (i = 2; i < gui_size_playq_height && pe_top != NULL; i++) {
		pe_top = playq_entry_next(pe_top);
		idx_top++;
	}
	
	if (pe_top == NULL) {
		pe_top = playq_entry_last();
		idx_top = playq_len - 1;
	}
	
	pe_selected = pe_top;
	idx_selected = idx_top;

	gui_playq_redraw();
	PLAYQ_UNLOCK;
}

void
gui_playq_cursor_top(void)
{
	PLAYQ_LOCK;
	pe_top = pe_selected = playq_entry_first();
	idx_top = idx_selected = (pe_top != NULL) ? 0 : -1;

	gui_playq_redraw();
	PLAYQ_UNLOCK;
}

void
gui_playq_cursor_bottom(void)
{
	PLAYQ_LOCK;
	pe_selected = playq_entry_last();
	idx_selected = playq_len - 1;
	gui_playq_cursor_adjust_bottom();

	gui_playq_redraw();
	PLAYQ_UNLOCK;
}

void
gui_playq_song_remove(void)
{
	PLAYQ_LOCK;
	if (pe_selected != NULL)
		playq_song_fast_remove(pe_selected, idx_selected);
	PLAYQ_UNLOCK;
}

void
gui_playq_song_add_before(struct vfs_ent *ve)
{
	PLAYQ_LOCK;
	if (idx_selected <= 0)
		/* List is empty or adding before the first node */
		playq_song_add_head(ve);
	else
		playq_song_fast_add_before(ve, pe_selected, idx_selected);
	PLAYQ_UNLOCK;
}

void
gui_playq_song_add_after(struct vfs_ent *ve)
{
	PLAYQ_LOCK;
	if (idx_selected < 0 || idx_selected == playq_len - 1)
		/* List is empty or appending after the last node */
		playq_song_add_tail(ve);
	else
		playq_song_fast_add_after(ve, pe_selected, idx_selected);
	PLAYQ_UNLOCK;
}

==> herrie-0.6.tar.gz: herrie-0.6/src/gui/gui_playq.c <==
/*
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */
/**
 * @file gui/gui_playq.c
 */

#include <glib.h>

#include <string.h>

#include <trans.h>
#include <gui/gui_internal.h>
#include <playq.h>

static WINDOW *win_playq;

static struct playq_entry *pe_top, *pe_selected;
static unsigned int idx_top, idx_selected;

void
gui_playq_init(void)
{
	win_playq = newwin(gui_size_playq_height, 0, 1, 0);
	wbkgdset(win_playq, COLOR_PAIR(GUI_COLOR_BLOCK));

	pe_top = pe_selected = NULL;
	idx_top = idx_selected = 0;
	gui_playq_refresh();
}

void
gui_playq_destroy(void)
{
	delwin(win_playq);
}

/**
 * @brief Scroll the viewport up if the cursor is too far up.
 */
static void
gui_playq_cursor_adjust_top(void)
{
	if (idx_top > idx_selected) {
		pe_top = pe_selected;
		idx_top = idx_selected;
	}
}

/**
 * @brief Scroll the viewport down if the cursor is too far down.
 */
static void
gui_playq_cursor_adjust_bottom(void)
{
	int i;
	unsigned int idx;
	struct playq_entry *pe;

	pe = pe_selected;
	idx = idx_selected;

	/* Don't deal with empty selections */
	if (pe == NULL)
		return;
	
	for (i = 1; i < gui_size_playq_height; i++) {
		pe = playq_entry_prev(pe);
		idx--;

		/* The top of the list is in screen */
		if ((pe == NULL) || (pe == pe_top))
			return;
	}

	/* We've gone too far */
	pe_top = pe;
	idx_top = idx;
}

/**
 * @brief Calculate what width the column with the index numbers should
 *        have, based on the amount of songs in the playlist.
 */
static int
gui_playq_idxcol_width(void)
{
	int len;
	unsigned int idx;

	/*
	 * Calculate the number of digits needed to print the number of
	 * entries
	 */
	for (len = 1, idx = playq_len; idx >= 10; len++, idx /= 10);

	return (len);
}

/**
 * @brief Fill playlist window with contents by acquiring the GUI lock
 *        and assuming that the playlist is already locked.
 */
static void
gui_playq_redraw(void)
{
	int i, idxw, idxmaxw;
	unsigned int idx;
	char num[16];
	struct playq_entry *pe;

	GUI_LOCK;
	werase(win_playq);

	pe = pe_top;
	idx = idx_top;
	idxmaxw = gui_playq_idxcol_width();
	for (i = 0; i < gui_size_playq_height; i++) {
		if (pe == NULL) {
			/* We've reached the bottom of the list */
			mvwaddch(win_playq, i, 0, ' ');
			wbkgdset(win_playq, COLOR_PAIR(GUI_COLOR_BLOCK));
			wclrtobot(win_playq);
			break;
		}

		if (pe == pe_selected) {
			if (gui_input_curfocus == GUI_FOCUS_PLAYQ)
				wbkgdset(win_playq,
				    COLOR_PAIR(GUI_COLOR_SELECT));
			else
				wbkgdset(win_playq,
				    COLOR_PAIR(GUI_COLOR_DESELECT));

			g_assert(idx == idx_selected);
		}

		/* Small whitespace on the left, or > when black & white */
		if ((pe == pe_selected) && !has_colors()) {
			if (gui_input_curfocus == GUI_FOCUS_PLAYQ)
				mvwaddch(win_playq, i, 0, '>');
			else
				mvwaddch(win_playq, i, 0, '-');
		} else {
			mvwaddch(win_playq, i, 0, ' ');
		}
		wclrtoeol(win_playq);

		idxw = snprintf(num, sizeof num, "%d.", idx);
		mvwaddstr(win_playq, i, 2 + idxmaxw - idxw, num);
		mvwaddstr(win_playq, i, idxmaxw + 3, pe->name);

		wbkgdset(win_playq, COLOR_PAIR(GUI_COLOR_BLOCK));

		pe = playq_entry_next(pe);
		idx++;
	}

	wrefresh(win_playq);
	GUI_UNLOCK;
}

#ifdef BUILD_GUI_RESIZE
void
gui_playq_resize(void)
{
	GUI_LOCK;
	wresize(win_playq, gui_size_playq_height, COLS);
	clearok(win_playq, TRUE);
	GUI_UNLOCK;
	
	PLAYQ_LOCK;
	gui_playq_cursor_adjust_bottom();
	gui_playq_redraw();
	PLAYQ_UNLOCK;
}
#endif /* BUILD_GUI_RESIZE */

void
gui_playq_refresh(void)
{
	PLAYQ_LOCK;
	gui_playq_redraw();
	PLAYQ_UNLOCK;
}

/**
 * @brief Adjust a reference with index to an entry in the playlist,
 *        right before another entry in the playlist gets removed.
 */
static void
gui_playq_adjust_pre_removal(struct playq_entry **pe,
    unsigned int *curidx, unsigned int newidx, int follow)
{
	struct playq_entry *pe_new;
	g_assert((*pe != NULL) && (*curidx > 0));

	if (*curidx > newidx) {
		/* Item got removed above us */
		pe_new = playq_entry_next(*pe);
		if (follow || pe_new == NULL) {
			/* Follow our selection */
			(*curidx)--;
		} else {
			/* Keep the current offset */
			*pe = pe_new;
		}
	} else if (*curidx == newidx) {
		/* Current item gets removed */
		pe_new = playq_entry_next(*pe);
		if (pe_new != NULL) {
			/* Stay at the same position */
			*pe = pe_new;
		} else {
			/* Cannot go down - go up */
			*pe = playq_entry_prev(*pe);
			(*curidx)--;
		}
	}

	g_assert((*pe == NULL) == (*curidx == 0));
}

void
gui_playq_notify_pre_removal(unsigned int index)
{
	gui_playq_adjust_pre_removal(&pe_top, &idx_top, index, 0);
	gui_playq_adjust_pre_removal(&pe_selected, &idx_selected, index, 1);
	gui_playq_cursor_adjust_top();
}

/**
 * @brief Adjust a reference with index to an entry in the playlist,
 *        right after a new entry in the playlist gets inserted.
 */
static void
gui_playq_adjust_post_insertion(struct playq_entry **pe,
    unsigned int *curidx, unsigned int newidx, int follow)
{
	g_assert((*pe == NULL) == (*curidx == 0));

	if (*curidx == 0) {
		/* List was empty - we take the first slot */
		*pe = playq_entry_first();
		*curidx = 1;
	} else if (*curidx >= newidx) {
		/* It happened above us */
		if (follow) {
			/* Follow our selection */
			(*curidx)++;
		} else {
			/* Keep the current offset */
			*pe = playq_entry_prev(*pe);
		}
	}

	g_assert((*pe != NULL) && (*curidx > 0));
}

void
gui_playq_notify_post_insertion(unsigned int index)
{
	gui_playq_adjust_post_insertion(&pe_top, &idx_top, index, 0);
	gui_playq_adjust_post_insertion(&pe_selected, &idx_selected, index, 1);
	gui_playq_cursor_adjust_bottom();
}

void
gui_playq_notify_post_randomization(void)
{
	unsigned int idx;

	/*
	 * After a randomization the indices do not match the
	 * corresponding objects. We could trace the objects and update
	 * the indices, but we might as well better keep the current
	 * indices and update the objects. This is more convenient for
	 * the user.
	 */
	if (idx_top > 0) {
		/* Top position */
		for (idx = 1, pe_top = playq_entry_first();
		    idx < idx_top;
		    idx++, pe_top = playq_entry_next(pe_top));
		/* Selection - just continue previous traversal */
		for (pe_selected = pe_top;
		    idx < idx_selected;
		    idx++, pe_selected = playq_entry_next(pe_selected));
	}
}

void
gui_playq_notify_done(void)
{
	gui_playq_redraw();
}

void
gui_playq_cursor_up(void)
{
	struct playq_entry *pe;

	PLAYQ_LOCK;
	if (pe_selected != NULL) {
		pe = playq_entry_prev(pe_selected);
		if (pe != NULL) {
			/* XXX */
			if (pe_top == pe_selected) {
				pe_top = pe;
				idx_top--;
			}
			
			pe_selected = pe;
			idx_selected--;

			gui_playq_redraw();
		} else {
			gui_msgbar_update(_("You are at the first song."), 0);
		}
	}
	PLAYQ_UNLOCK;
}

void
gui_playq_cursor_down(void)
{
	struct playq_entry *pe;

	PLAYQ_LOCK;
	if (pe_selected != NULL) {
		pe = playq_entry_next(pe_selected);
		if (pe != NULL) {
			pe_selected = pe;
			idx_selected++;

			/*
			 * Make sure the cursoer doesn't leave the
			 * bottom of the screen
			 */
			gui_playq_cursor_adjust_bottom();
			gui_playq_redraw();
		} else {
			gui_msgbar_update(_("You are at the last song."), 0);
		}
	}
	PLAYQ_UNLOCK;
}

void
gui_playq_cursor_pageup(void)
{
	int i;

	PLAYQ_LOCK;
	if (idx_top <= 1)
		goto done;

	/* Cursor = bottom of the screen */
	pe_selected = playq_entry_next(pe_top);
	idx_selected = idx_top + 1;

	/* Adjust the viewport */
	for (i = 2; i < gui_size_playq_height && pe_top != NULL; i++) {
		pe_top = playq_entry_prev(pe_top);
		idx_top--;
	}
	if (pe_top == NULL) {
		pe_top = playq_entry_first();
		idx_top = 1;
	}
	
	/* Original screen may have had one item */
	if (pe_selected == NULL) {
		pe_selected = pe_top;
		idx_selected = idx_top;
	}
	
	gui_playq_redraw();
done:
	PLAYQ_UNLOCK;
}

void
gui_playq_cursor_pagedown(void)
{
	int i;

	PLAYQ_LOCK;
	/* A la vi: page down minus two */
	for (i = 2; i < gui_size_playq_height && pe_top != NULL; i++) {
		pe_top = playq_entry_next(pe_top);
		idx_top++;
	}
	
	if (pe_top == NULL) {
		pe_top = playq_entry_last();
		idx_top = playq_len;
	}
	
	pe_selected = pe_top;
	idx_selected = idx_top;

	gui_playq_redraw();
	PLAYQ_UNLOCK;
}

void
gui_playq_cursor_top(void)
{
	PLAYQ_LOCK;
	pe_top = pe_selected = playq_entry_first();
	idx_top = idx_selected = (pe_top != NULL) ? 1 : 0;

	gui_playq_redraw();
	PLAYQ_UNLOCK;
}

void
gui_playq_cursor_bottom(void)
{
	PLAYQ_LOCK;
	pe_selected = playq_entry_last();
	idx_selected = playq_len;
	gui_playq_cursor_adjust_bottom();

	gui_playq_redraw();
	PLAYQ_UNLOCK;
}

void
gui_playq_song_remove(void)
{
	PLAYQ_LOCK;
	if (pe_selected != NULL)
		playq_song_fast_remove(pe_selected, idx_selected);
	else
		gui_msgbar_update(_("There are no songs."), 0);
	PLAYQ_UNLOCK;
}

void
gui_playq_song_remove_all(void)
{
	/* Don't care about locking here */
	if (playq_len == 0) {
		gui_msgbar_update(_("There are no songs."), 0);
		return;
	}

	if (gui_input_askyesno(_("Remove all songs from the playlist?")) == 0)
		playq_song_remove_all();
}

void
gui_playq_song_randomize(void)
{
	/* Don't care about locking here */
	if (playq_len == 0) {
		gui_msgbar_update(_("There are no songs."), 0);
		return;
	}

	if (gui_input_askyesno(_("Randomize the playlist?")) == 0)
		playq_song_randomize();
}

void
gui_playq_song_add_before(struct vfs_ent *ve)
{
	PLAYQ_LOCK;
	if (idx_selected <= 1)
		/* List is empty or adding before the first node */
		playq_song_add_head(ve);
	else
		playq_song_fast_add_before(ve, pe_selected, idx_selected);
	PLAYQ_UNLOCK;
}

void
gui_playq_song_add_after(struct vfs_ent *ve)
{
	PLAYQ_LOCK;
	if (idx_selected == 0 || idx_selected == playq_len)
		/* List is empty or appending after the last node */
		playq_song_add_tail(ve);
	else
		playq_song_fast_add_after(ve, pe_selected, idx_selected);
	PLAYQ_UNLOCK;
}

void
gui_playq_song_moveup(void)
{
	PLAYQ_LOCK;
	if (idx_selected == 0)
		gui_msgbar_update(_("There are no songs."), 0);
	else if (idx_selected == 1)
		gui_msgbar_update(_("The song is already at the top "
		    "of the playlist."), 0);
	else
		playq_song_fast_moveup(pe_selected, idx_selected);
	PLAYQ_UNLOCK;
}

void
gui_playq_song_movedown(void)
{
	PLAYQ_LOCK;
	if (idx_selected == 0)
		gui_msgbar_update(_("There are no songs."), 0);
	else if (idx_selected == playq_len)
		gui_msgbar_update(_("The song is already at the bottom "
		    "of the playlist."), 0);
	else
		playq_song_fast_movedown(pe_selected, idx_selected);
	PLAYQ_UNLOCK;
}

void
gui_playq_searchnext(void)
{
	struct playq_entry *pe;
	unsigned int idx;

	if (gui_input_cursearch == NULL) {
		/* Nothing to search for */
		gui_msgbar_update(_("No search pattern."), 0);
		return;
	}

	PLAYQ_LOCK;

	if (pe_selected == NULL) {
		/* Playlist is empty */
		gui_msgbar_update(_("There are no songs."), 0);
		goto out;
	}

	/* Step 1: search from the next entry to the bottom */
	for (pe = playq_entry_next(pe_selected), idx = idx_selected + 1;
	    pe != NULL; pe = playq_entry_next(pe), idx++) {
		if (strcasestr(pe->name, gui_input_cursearch) != NULL)
			break;
	}
	if (pe != NULL) {
		/* We've found an item below us */
		pe_selected = pe;
		idx_selected = idx;

		/* Item may be below the viewport */
		gui_playq_cursor_adjust_bottom();
		gui_playq_redraw();
		goto out;
	}

	/* Step 2: search from beginning next entry to the bottom */
	for (pe = playq_entry_first(), idx = 1;
	    pe != playq_entry_next(pe_selected);
	    pe = playq_entry_next(pe), idx++) {
		if (strcasestr(pe->name, gui_input_cursearch) != NULL)
			break;
	}
	if (pe != playq_entry_next(pe_selected)) {
		/* We've found an item above us */
		pe_selected = pe;
		idx_selected = idx;

		/* Item may be above the viewport */
		gui_playq_cursor_adjust_top();
		gui_playq_redraw();

		/* Also show the wrapping message */
		gui_msgbar_update(_("Search wrapped to top."), 0);
		goto out;
	}

	gui_msgbar_update(_("Not found."), 0);
out:
	PLAYQ_UNLOCK;
}
